@@ -7,11 +7,10 @@ import datetime
 import subprocess
 import platform
 import functools
-from typing import List, Any
+from typing import List, Any, Tuple
 
 from parent_class import ParentClass
-from py_starter import py_starter as ps
-
+import py_starter as ps
 
 DELIM = '/'
 SECONDARY_DELIM = '\\'
@@ -70,27 +69,6 @@ def convert_bytes( bytes: int, conversion: str = 'MB' ) -> Tuple[ Any, Any ]:
 
     return bytes, None
 
-def walk_root_dir( path ):
-
-    roots = []
-    dirs = []
-    files = []
-
-    for i,j,k in os.walk(path):
-
-        roots.append(i)
-        dirs.append(j)
-        files.append(k)
-
-    return roots, dirs, files
-
-def rename_files_to_new_directory( input_folder, export_folder ):
-
-    full_input_paths = get_paths_with_common_parent( input_folder )
-    relative_input_paths = remove_prefix_from_paths( input_folder, full_input_paths )
-    full_export_paths = add_prefix_to_paths( export_folder, relative_input_paths )
-
-    return full_input_paths, full_export_paths
 
 def add_prefix_to_paths( prefix_path, relative_paths ):
 
@@ -118,22 +96,6 @@ def remove_prefix_from_paths( prefix_path, full_paths ):
 
     return relative_paths
 
-def get_paths_with_common_parent( path ):
-
-    """Returns full paths with common parent path
-     Note: does not return empty directories"""
-
-    roots, dirs, files = walk_root_dir( path )
-    full_paths = []
-
-    for i in range(len(roots)):
-        for file in files[i]:
-
-            full_path = join( roots[i], file )
-            full_paths.append( full_path )
-
-    return full_paths
-
 def remove_hanging_slashes( path ):
 
     """removes ending slashes from paths
@@ -279,6 +241,14 @@ class Dir (ParentClass) :
         """add more dirs to the Dir path"""
         return join( dir, *other_dirs )
 
+    def join_Dir( self, other_Dir: Dir ) -> Dir:
+
+        return Dir( self.join( other_Dir.path ) )
+
+    def join_Path( self, other_Path: Path ) -> Path:
+
+        return Path( self.join( other_Path.path ) )
+
     @instance_method
     def open( self, *args, **kwargs ):
         pass
@@ -289,19 +259,10 @@ class Dir (ParentClass) :
         """Opens the dir in the file explorer """
 
         if platform.system() == 'Windows':
-            os.startfile(path)
+            os.startfile(dir)
 
         elif platform.system() == 'Darwin':
-            subprocess.call(['open', path])
-
-    @instance_method
-    def get_unique_filename( self, *args, **kwargs ):
-        pass
-
-    def get_unique_filename_dir( self, filename ):
-
-        if self.type_dir:
-            return get_unique_filename( filename, export_dir = self.p )
+            subprocess.call(['open', dir])
 
     @instance_method
     def exists( self, *args, **kwargs ):
@@ -393,6 +354,16 @@ class Dir (ParentClass) :
         """returns all directories and files contained in dir""" 
         return os.listdir( dir )
 
+    def get_rel( self, other_Dir ) -> Dir:
+
+        """Given a Dir object, find the relative Dir from Dir to self"""
+
+        return Dir( Dir.get_rel_dir( self.path, other_Dir.path ) )
+
+    @staticmethod
+    def get_rel_dir( dir, other_dir ) -> str:
+        return os.path.relpath( dir, other_dir )
+
     def list_contents_Paths( self, block_dirs: bool = True, block_paths: bool = False ) -> Paths: 
 
         """returns all first sublevel contents of Directory as a Paths instance"""
@@ -413,20 +384,41 @@ class Dir (ParentClass) :
             
         return Paths_inst
 
-    def walk_contents_Paths( self, block_dirs: bool = True, block_paths: bool = False, folders_to_skip: List[str] = [] ) -> Paths:
+    def walk( self, folders_to_skip: List[str] = ['.git'] ) -> Paths:
+
+        """Walk through all the contents of the directory"""
+
+        Paths_inst = Paths()
+        Paths_inst._add( self )
+
+        Paths_under = self.list_contents_Paths( block_dirs = False, block_paths = False )
+        for Path_inst in Paths_under:
+
+            if Path.is_Path( Path_inst ):
+                Paths_inst._add( Path_inst )
+
+            elif Path.is_Dir( Path_inst ):
+                
+                if Path_inst.dirs[-1] not in folders_to_skip:
+                    Paths_inst.merge( Path_inst.walk( folders_to_skip = folders_to_skip ) )
+
+        return Paths_inst
+
+    def walk_contents_Paths( self, block_dirs: bool = True, block_paths: bool = False, folders_to_skip: List[str] = ['.git'] ) -> Paths:
 
         """get all Paths and/or Dirs underneath the entire directory, optional params for returning paths and/or dirs"""
 
-        Paths_inst = Paths()
-        subfolders = get_subfolders_under_parent( dir, folders_to_skip = folders_to_skip )
-        for subfolder in subfolders:
-            Paths_under_subfolder = list_contents_Paths( subfolder, block_dirs = block_dirs, block_paths = block_paths )
+        Paths_inst = self.walk( folders_to_skip=folders_to_skip )
+        keep_Paths = Paths()
 
-            for Path_under_subfolder in Paths_under_subfolder:
-                if not is_dir(Path_under_subfolder.p) and Path_under_subfolder.dirs[-1] not in folders_to_skip:
-                    Paths.append( Path_under_subfolder )
+        for Path_inst in Paths_inst:
+            if Path_inst.type_path and not block_paths:
+                keep_Paths._add( Path_inst )
 
-        return Paths
+            if Path_inst.type_dir and not block_dirs:
+                keep_Paths._add( Path_inst )
+
+        return keep_Paths
 
     def get_child_Dirs( self, folders_to_skip: List[str] = ['.git'] ) -> Dirs:
                 
@@ -530,6 +522,7 @@ class Path( Dir ):
         """copies an the contents from source to destination"""
 
         if destination_Path != None:
+            destination_Path.create_parents()
             destination_path = destination_Path.path
 
         if print_off:
@@ -561,7 +554,7 @@ class Path( Dir ):
             inp = input('Type "delete" to delete ' + str(path) + ': ')
 
         if inp == 'delete':
-            if path_exists(path):
+            if Path.exists_path(path):
                 if print_off:
                     print ('Deleting file ' + str(path))
                 try:
@@ -597,30 +590,20 @@ class Path( Dir ):
         else:
             return False
 
-    @instance_method
-    def get_size( self, *args, **kwargs ):
-        pass
-
-    @staticmethod
-    def get_size_path( path, **kwargs ) -> None:
+    def get_size( self, **kwargs ) -> None:
 
         """get the size of the path"""
 
-        size = os.stat( path ).st_size
+        size = os.stat( self.path ).st_size
         converted_size, conversion = convert_bytes( size, **kwargs )
 
         self.size = converted_size
         self.size_units = conversion
 
-
-    @instance_method
-    def get_mtime( self, *args, **kwargs ):
-        pass
-
-    @staticmethod
-    def get_mtime_path( path ):
+    def get_mtime( self, *args, **kwargs ) -> datetime.datetime:
 
         """get the time of modification as a datetime object"""
+
         mtime = pathlib.Path(self.p).stat().st_mtime
         self.mtime = datetime.datetime.fromtimestamp( mtime )
         return self.mtime
@@ -664,7 +647,34 @@ class Path( Dir ):
 
         """imports the contents of path as a module"""
         return ps.import_module_from_path( path, **kwargs )
-    
+
+    @instance_method
+    def smart_format( self, *args, **kwargs) -> str:
+        pass
+
+    @staticmethod
+    def smart_format_path( path, formatting_dict, write = True, **kwargs ) -> str:
+
+        """passthrough function for py_starter.smart_format()"""
+        string = Path.read_path( path )
+        formatted_string = ps.smart_format( string, formatting_dict, **kwargs )
+
+        if write:
+            Path.write_path( path, string = formatted_string )
+
+        return formatted_string
+
+
+    def get_rel( self, Dir_inst ) -> Path:
+
+        """Given a Dir object, find the relative Path from Dir to self"""
+
+        return Path( Path.get_rel_path( self.path, Dir_inst.path ) )
+
+    @staticmethod
+    def get_rel_path( path, dir ) -> str:
+        return os.path.relpath( path, dir )
+
     @staticmethod
     def get_filename( path: str ) -> str:
 
@@ -767,6 +777,22 @@ class Dirs(ParentClass):
         self.len_Dirs = len(self)
         return self._print_one_line_atts_helper( atts = ['type','len_Dirs'], print_off = print_off, leading_string = leading_string )
 
+    def join_Dir( self, Dir_inst: Dir ) -> Paths:
+
+        """Joins Dir_inst to each Dir/Path contained in the Object"""
+
+        Paths_inst = Paths()
+
+        for DirPath in self:
+
+            if DirPath.type_path:
+                Paths_inst._add( Dir_inst.join_Path( DirPath ) )
+            
+            if DirPath.type_dir:
+                Paths_inst._add( Dir_inst.join_Dir( DirPath ) )
+
+        return Paths_inst
+
     def merge( self, other_Dirs: Any ):
 
         """add all Dir objects from another Dirs instance to self"""
@@ -817,3 +843,16 @@ class Paths( Dirs ):
         self.len_Paths = len(self)
         return self._print_one_line_atts_helper( atts = ['type','len_Paths'], print_off = print_off, leading_string = leading_string )
 
+
+    def get_rels( self, Dir_inst ) -> Paths:
+
+        """Given a Dir object, find the relative Paths from Dir to the Paths"""
+
+        Paths_inst = Paths()
+        for P in self:
+            Paths_inst._add( P.get_rel( Dir_inst ) )
+
+        return Paths_inst
+
+   
+
